/* Copyright (c) 2015-2016 Dovecot authors, see the included COPYING file */

#ifndef PUSH_NOTIFICATION_EVENT_MAILBOXRENAME_H
#define PUSH_NOTIFICATION_EVENT_MAILBOXRENAME_H


struct push_notification_event_mailboxrename_data {
    const char *old_mbox;
};


#endif /* PUSH_NOTIFICATION_EVENT_MAILBOXRENAME_H */

