#ifndef IMAP_RESP_CODE_H
#define IMAP_RESP_CODE_H

/* IMAP response codes (RFC 5530) */
#define IMAP_RESP_CODE_UNAVAILABLE "UNAVAILABLE"
#define IMAP_RESP_CODE_AUTHFAILED "AUTHENTICATIONFAILED"
#define IMAP_RESP_CODE_AUTHZFAILED "AUTHORIZATIONFAILED"
#define IMAP_RESP_CODE_EXPIRED "EXPIRED"
#define IMAP_RESP_CODE_PRIVACYREQUIRED "PRIVACYREQUIRED"
#define IMAP_RESP_CODE_CONTACTADMIN "CONTACTADMIN"
#define IMAP_RESP_CODE_NOPERM "NOPERM"
#define IMAP_RESP_CODE_INUSE "INUSE"
#define IMAP_RESP_CODE_EXPUNGEISSUED "EXPUNGEISSUED"
#define IMAP_RESP_CODE_CORRUPTION "CORRUPTION"
#define IMAP_RESP_CODE_SERVERBUG "SERVERBUG"
#define IMAP_RESP_CODE_CLIENTBUG "CLIENTBUG"
#define IMAP_RESP_CODE_CANNOT "CANNOT"
#define IMAP_RESP_CODE_LIMIT "LIMIT"
#define IMAP_RESP_CODE_OVERQUOTA "OVERQUOTA"
#define IMAP_RESP_CODE_ALREADYEXISTS "ALREADYEXISTS"
#define IMAP_RESP_CODE_NONEXISTENT "NONEXISTENT"

#define IMAP_RESP_CODE_UNKNOWN_CTE "UNKNOWN-CTE" /* BINARY */

#endif
